;;; Repertoire: ~/LeShirka/LeIvan/CodeIvan/Requete/
;
;     +----------------------------------------+-----------------------------------------+
;     | fichier :  ivan-re-execut.ll           | package : #:ivan:re-interf              |
;     +----------------------------------------+-----------------------------------------+
;     |                                                                                  |
;     |                     Projet de memoire CNAM en informatique                       |
;     |             Interface de Visualisation et d'Aide a la Navigation                 |
;     |                dans une base de connaissances a objets Shirka.                   |
;     |                        Sylvain Grivaud          1991                             |
;     |                                                                                  |
;     +----------------------------------------------------------------------------------+
;     |                                                                                  |
;     | Ce fichier contient les fonctions d'evaluation d'une requete concue par le       |
;     | module des requetes                                                              |
;     |                                                                                  |
;     +----------------------------------------------------------------------------------+

;;;*****************************************************************************************

(defvar #:sys-package:colon '#:ivan:re-interf)

;;;*****************************************************************************************
;;; en entree   
;;;     list-attr       : listes des attributs (liste de vecteurs)
;;;     list-comp       : liste des comparateurs
;;;     list-val        : liste des valeurs cherchees
;;;     list-conn       : liste des connecteurs
;;;     list-depart     : liste des instances de depart prises dans le
;;;                       selectionneur des instances
;;; en sortie :
;;;     la liste des instances extraient de list-depart qui satisfont
;;;     la requete

(de :evaluer-requete (list-attr
                             list-comp
                             list-val
                             list-conn
                             list-depart)
    
    (let
         (
          ; liste renvoyee : egale a la liste de depart par defaut
          ;                  construite dynamiquement par evaluation
          ;                  successive des criteres
          (list-resultat list-depart)
          ; liste des connecteurs logiques : inserer un "et" en tete de liste
          ; pour avoir le meme nombre d'attributs (list-attr) et de connecteurs
          ; (list-cl).
          (list-cl       (cons #M#:txt-const:et list-conn))
          ; compteur pour extraire le critere courant
          (n 0)
          )
         ; boucle principale : extraire chaque connecteur logique
         (while list-cl
                
                (let
                     (
                      (conn (car   list-cl))   ; le connecteur
                      (attr (nth n list-attr)) ; l'attribut : vecteur ou liste de vecteurs
                      (comp (nth n list-comp)) ; le comparateur sous sa forme externe
                      (val  (nth n list-val))  ; la valeur a chercher
                      )
                     
                     (cond
                           ;
                           ; cas du ET :
                           ;     la liste d'entree est la liste resultat de la 
                           ;     derniere evaluation (la premiere fois c'est donc
                           ;     liste de depart)
                           ((equal conn #M#:txt-const:et)
                            (setq list-resultat (:evaluer-crit attr
                                                        comp
                                                        val
                                                        list-resultat
                                                        ))
                            )
                           ;
                           ; cas du OU :
                           ;     la liste d'entree  est la liste de depart;
                           ;     la liste de sortie est l'union entre la liste resultat
                           ;     de l'evaluation precedente et la liste resultat de
                           ;     de l'evaluation du ou.
                           
                           ((equal conn #M#:txt-const:ou)
                            (setq list-resultat (union
                                                       list-resultat
                                                       (:evaluer-crit attr
                                                               comp
                                                               val
                                                               list-depart
                                                               )))
                            ) 
                           ;
                           ; cas du SAUF :
                           ;     la liste d'entree  est la liste resultat de la
                           ;                            derniere evaluation.
                           ;     la liste de sortie est la difference entre la
                           ;                            liste resultat de l'evaluation
                           ;                            precedente et la liste resultat
                           ;                            de l'evaluation du sauf.
                           
                           ((equal conn #M#:txt-const:sauf)
                            (setq list-resultat (set-difference
                                                        list-resultat
                                                        (:evaluer-crit attr
                                                                comp
                                                                val
                                                                list-resultat
                                                                )))
                            ) 
                           ;
                           ; cas d'erreur
                           ;
                           ( t
			    (when #:trace:re
                              (print "erreur !!! conn " conn " # " (type-of conn))
			    )
                           )
                           )
                     )
                ; passer au suivant
                (setq n (+ n 1)) 
                (nextl list-cl)
                )
         list-resultat
         )
    )
;;;*****************************************************************************************
;;; fonction d'evaluation d'un critere
;;; attr        : attribut = 1 vecteur ou une liste de vecteurs
;;; comp        : comparateur
;;; vr          : valeur de recherche
;;; list-in     : liste en entree
;;; list-out    : liste des instances satisfaisant le critere

(de :evaluer-crit (attr comp vr list-in)
    (let
         (   ; transformer attr en 1 liste
            (list-attr (if (listp attr)
                           attr
                           (list attr) ))
            list-out
            )
         (setq list-out
               ; si le CDR de attr est nil, alors le CAR de attr est
               ; le vecteur d'un attribut terminal d'un type simple
               ; c'est sur lui qu'il faut effectuer la recherche
               (if (null (cdr list-attr))
                   ;  c'est un attribut simple
                   (:evaluer-crit-simp (car list-attr) comp vr list-in)
                   ; sinon c'est une requete sur un attribut complexe
                   (:evaluer-crit-comp attr comp vr list-in)
                   )
               )
         list-out
         )
    )
;;;*****************************************************************************************
;;; fonction d'evaluation d'un critere complexe 
;;; attr        : attribut = 1 vecteur ou une liste de vecteurs
;;; comp        : comparateur
;;; vr          : valeur de recherche
;;; list-in     : liste en entree
;;; renvoyer directement la liste des instances satisfaisant le critere

(de :evaluer-crit-comp (attr comp vr list-in)
    (let
         (
          ; le CAR de attr contient un vecteur [attribut classe type]
          ; extraire le nom de l'attribut        
          (nom-attr (#:ivan:ed-util:symbole (vref (car attr) 0)))
          ; valeur de l'attribut nom-attr
          l-val
          ; l-val est transforme en liste (va = valeurs attributs)
          va
          )
         (mapcan '(lambda (nom-inst)
                          ; je cherche la valeur de l'attribut nom-attr
                          (setq l-val (#:shirka:moteur:valeur? nom-attr nom-inst)          
                                va    (if (listp l-val) l-val (list l-val))
                                )
                          ; appel de :evaluer-crit sur la suite de la liste attr
                          (when (:evaluer-crit (cdr attr) comp vr va)
                                ; l'instance nom-inst est satisfaisante
                                (list nom-inst)
                                )
                          )
                 list-in
                 )
         )
    )
;;;*****************************************************************************************
;;; fonction d'evaluation d'un critere simple 
;;; attr        = vecteur de l'attribut [attribut classe type]
;;; comp        = comparateur sous sa forme externe
;;; vr          = valeur de recherche
;;; list-in     = liste des instances en entree

(de :evaluer-crit-simp (attr comp vr list-in)
    (let
         (
          ; extraire le nom de l'attribut du vecteur
          (nom-attr  (#:ivan:ed-util:symbole (vref attr 0)))
          ; extraire le type de l'attribut
          (type-attr (#:ivan:ed-util:symbole (vref attr 2)))
          ; liste des noms d'attributs de l'instance courante
          list-attr
          ; valeur de l'attribut de l'instance courante
          va
          ; forme lisp du critere a evaluer
          (crit-lisp (:const-crit-lisp vr comp))
          )
         
         (mapcan '(lambda (nom-inst)
                          ; construire la liste des attributs de l'instance
                          (setq
                                list-attr (#:shirka:noyau:l-nom-att-he-class
                                                  (#:shirka:noyau:classe-de nom-inst)
                                                  nil nil)
                                )
                          ; verifier que l'attribut nom-attr est un attribut
                          ; de l'instance avant de chercher sa valeur
                          (when (member nom-attr list-attr)
                                ; demander a Shirka la valeur de l'attribut (avec inference)
                                (setq
                                      va    (#:shirka:moteur:valeur? nom-attr nom-inst)
                                      )
                                ; evaluer le critere LISP sur l'argument va
                                ; si vrai alors retourner l'instance
                                (when (funcall crit-lisp va)
                                      (list nom-inst)
                                      )
                                )
                          )
                 list-in
                 )
         )
    )
;;;*****************************************************************************************
;;; fonction de construction du critere simple
;;; vr   : valeur requete
;;; comp : comparateur

(de :const-crit-lisp (vr comp)
    (let
         (
          l-va
          l-vr
          (comp-cadre (format () "~17@A" comp)) ; comparateur cadre sur 17 caracteres
          )
         (cond
               ;------------------------------------------------
               ( (equal comp-cadre #M#:txt-const:text-egal)
                (setq l-vr (if (listp vr) (mapcar 'string vr) (string vr)) )
                `(lambda (x)
                         (let
                              (
                               (l-va (if (listp x) (mapcar 'string x) (string x)) )
                               )
                              (equal l-va (quote ,l-vr))
                              ))
                )     
               ;------------------------------------------------
               ( (equal comp-cadre  #M#:txt-const:text-diff)
                (setq l-vr (if (listp vr) (mapcar 'string vr) (string vr)) )
                `(lambda (x)
                         (let
                              (
                               (l-va (if (listp x) (mapcar 'string x) (string x)) )
                               )
                              (nequal l-va (quote ,l-vr))
                              ))
                )     
               ;------------------------------------------------
               ( (equal comp-cadre  #M#:txt-const:text-sup)
                (lambda (x)
                        (if x
                            (> x vr)
                            nil
                            )
                        )
                )     
               
               ;------------------------------------------------
               ( (equal comp-cadre  #M#:txt-const:text-inf)
                (lambda (x)
                        (if x
                            (< x vr)
                            nil
                            )
                        )
                )
               ;------------------------------------------------
               ( (equal comp-cadre #M#:txt-const:text-supegal)
                (lambda (x)
                        (if x
                            (>= x vr)
                            nil
                            )
                        )
                )
               ;------------------------------------------------
               ( (equal comp-cadre #M#:txt-const:text-infegal)
                (lambda (x)
                        (if x
                            (<= x vr)
                            nil
                            )
                        )
                )
               ;------------------------------------------------
               ( (equal comp-cadre #M#:txt-const:text-cont)
                (setq l-vr (mapcar 'string vr) )
                
                `(lambda (x)
                         (let
                              (
                               (l-va (mapcar 'string x) )
                               )
                              (subsetp (quote ,l-vr) l-va 'equal)
                              ))
                )
               ;------------------------------------------------
               ( (equal comp-cadre #M#:txt-const:text-ncont)
                (setq l-vr (mapcar 'string vr) )
                `(lambda (x)
                         (let
                              (
                               (l-va (mapcar 'string x) )
                               )
                              (not (subsetp (quote ,l-vr) l-va 'equal))
                              ))
                )
               ;------------------------------------------------
               ( (equal comp-cadre #M#:txt-const:text-deb)
                (lambda (x)
                        (equal (index (string vr) (string x) 0) 0)
                        )
                )
               ;------------------------------------------------
               ( (equal comp-cadre #M#:txt-const:text-fin)
                (lambda (x)
                        (equal 
                               (index (string vr) (string x)  (- (plength (string x))
                                                                 (plength (string vr)) ) )
                               
                               (- (plength (string x))
                                  (plength (string vr)) )
                               )
                        )
                )
               ;------------------------------------------------
               ( (equal comp-cadre #M#:txt-const:text-incl)
                (lambda (x)
                        (index (string vr) (string x) 0)
                        )
                )
               ;------------------------------------------------
               ( (equal comp-cadre #M#:txt-const:text-appart)
                (setq l-vr (mapcar 'string vr) )
                `(lambda (x)
                         (lets
                               (
                                (l-x  (if (listp x) x (list x))) ; la valeur de l'attribut n'est 
                                ; pas forcement une liste
                                (l-va (mapcar 'string l-x) )
                                )
                               (if l-va
                                   (subsetp l-va (quote ,l-vr) 'equal)
                                   nil   ; cas ou l'attribut n'a pas de valeur 
                                   ; le test resoud le pb de (subsetp nil '(a b))
                                   ; qui est vrai pour le-lisp !
                                   )
                               ))
                )
               ;------------------------------------------------
               ( t
		(when #:trace:re
                  (print "erreur !!! comp " comp " # " (type-of comp))
		)
               )
               ;------------------------------------------------
               )
         )
    )
